Package: uwot
Title: The Uniform Manifold Approximation and Projection (UMAP) Method
        for Dimensionality Reduction
Version: 0.2.5
Authors@R: c(
    person("James", "Melville", , "jlmelville@gmail.com", role = c("aut", "cre", "cph")),
    person("Aaron", "Lun", role = "ctb"),
    person("Mohamed Nadhir", "Djekidel", role = "ctb"),
    person("Yuhan", "Hao", role = "ctb"),
    person("Dirk", "Eddelbuettel", role = "ctb"),
    person("Wouter", "van der Bijl", role = "ctb"),
    person("Hugo", "Gruson", role = "ctb")
  )
Description: An implementation of the Uniform Manifold Approximation and
    Projection dimensionality reduction by McInnes et al. (2018)
    <doi:10.48550/arXiv.1802.03426>. It also provides means to transform new data and
    to carry out supervised dimensionality reduction. An implementation of
    the related LargeVis method of Tang et al. (2016) <doi:10.48550/arXiv.1602.00370>
    is also provided. This is a complete re-implementation in R (and C++,
    via the 'Rcpp' package): no Python installation is required. See the
    uwot website (<https://github.com/jlmelville/uwot>) for more
    documentation and examples.
License: GPL (>= 3)
SystemRequirements: C++17
URL: https://github.com/jlmelville/uwot,
        https://jlmelville.github.io/uwot/
BugReports: https://github.com/jlmelville/uwot/issues
Depends: Matrix
Imports: FNN, irlba, methods, Rcpp, RcppAnnoy (>= 0.0.17), RSpectra
Suggests: bigstatsr, covr, knitr, RcppHNSW, rmarkdown, rnndescent,
        testthat
LinkingTo: dqrng, Rcpp, RcppAnnoy, RcppProgress
VignetteBuilder: knitr
Config/Needs/website: rmarkdown
Encoding: UTF-8
Config/roxygen2/version: 8.1.0
NeedsCompilation: yes
Packaged: 2026-08-28 19:36:59 UTC; jlmel
Author: James Melville [aut, cre, cph],
  Aaron Lun [ctb],
  Mohamed Nadhir Djekidel [ctb],
  Yuhan Hao [ctb],
  Dirk Eddelbuettel [ctb],
  Wouter van der Bijl [ctb],
  Hugo Gruson [ctb]
Maintainer: James Melville <jlmelville@gmail.com>
Repository: CRAN
Date/Publication: 2026-08-29 13:10:02 UTC
Built: R 4.6.1; aarch64-apple-darwin23; 2026-08-29 14:47:23 UTC; unix
Archs: uwot.so.dSYM
