# Generated by roxygen2: do not edit by hand

export(load_uwot)
export(lvish)
export(optimize_graph_layout)
export(save_uwot)
export(similarity_graph)
export(simplicial_set_intersect)
export(simplicial_set_union)
export(tumap)
export(umap)
export(umap2)
export(umap_transform)
export(unload_uwot)
import(Matrix)
importFrom(Rcpp,sourceCpp)
importFrom(methods,new)
useDynLib(uwot, .registration=TRUE)
